# -*- coding: utf-8 -*-
"""
Interactive tool to draw mask on an image or image-like array.

TODO:
    * need concept of subannotation
    * need to take options on a right click of an annotation
    * add support for arbitrary polygons back in .
    * rename species_list to label_list or category_list
    * Just use metadata instead of species / category / label
    # Need to incorporate parts into metadata

Notes:
    3. Change bounding box and update continuously to the original image the
    new ANNOTATIONs

    2. Make new window and frames inside, double click to pull up normal window
    with editing start with just taking in 6 images and ANNOTATIONs

    1. ANNOTATION ID number, then list of 4 tuples

    python -m utool.util_inspect check_module_usage --pat="interact_annotations.py"

References:
    Adapted from matplotlib/examples/event_handling/poly_editor.py
    Jan 9 2014: taken from: https://gist.github.com/tonysyu/3090704

CommandLine:
    python -m wbia.plottool.interact_annotations --test-test_interact_annots --show
"""
import logging
import six
import re
import numpy as np

try:
    import vtool as vt
except ImportError:
    pass
import utool as ut
import itertools as it
import matplotlib as mpl
from six.moves import zip, range
from . import draw_func2 as df2
from wbia.plottool import abstract_interaction

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


DEFAULT_SPECIES_TAG = '____'
# FIXE THESE TO BE GENERIC
ACCEPT_SAVE_HOTKEY = None  # 'ctrl+a'
ADD_RECTANGLE_HOTKEY = 'ctrl+a'  # 'ctrl+d'
ADD_RECTANGLE_FULL_HOTKEY = 'ctrl+f'
DEL_RECTANGLE_HOTKEY = 'ctrl+d'  # 'ctrl+r'
TOGGLE_LABEL_HOTKEY = 'ctrl+t'

HACK_OFF_SPECIES_TYPING = True
if HACK_OFF_SPECIES_TYPING:
    ADD_RECTANGLE_HOTKEY = 'a'  # 'ctrl+d'
    ADD_RECTANGLE_FULL_HOTKEY = 'f'
    DEL_RECTANGLE_HOTKEY = 'd'  # 'ctrl+r'
    TOGGLE_LABEL_HOTKEY = 't'

NEXT_IMAGE_HOTKEYS = ['right', 'pagedown']
PREV_IMAGE_HOTKEYS = ['left', 'pageup']

TAU = np.pi * 2


class AnnotPoly(mpl.patches.Polygon, ut.NiceRepr):
    """
    Helper to represent an annotation polygon
    wbia --aidcmd='Interact image' --aid=1

    Example:
        >>> # DISABLE_DOCTEST
        >>> from wbia.plottool.interact_annotations import *  # NOQA
        >>> verts = vt.verts_from_bbox([0, 0, 10, 10])
        >>> poly = AnnotPoly(None, 0, verts, 0, '____')
    """

    def __init__(
        poly,
        ax,
        num,
        verts,
        theta,
        species,
        fc=(0, 0, 0),
        line_color=(1, 1, 1),
        line_width=4,
        is_orig=False,
        metadata=None,
        valid_species=None,
        manager=None,
    ):

        super(AnnotPoly, poly).__init__(verts, animated=True, fc=fc, ec='none', alpha=0)
        poly.manager = manager
        # Ensure basecoords consistency
        poly.basecoords = vt.verts_from_bbox(vt.bbox_from_verts(poly.xy))
        # poly.basecoords = poly.xy
        poly.num = num
        poly.is_orig = is_orig
        poly.theta = theta
        poly.metadata = metadata
        poly.valid_species = valid_species
        poly.tab_list = valid_species
        # put in previous text and tabcomplete list for autocompletion
        poly.tctext = ''
        poly.tcindex = 0
        poly.anchor_idx = 2
        poly.child_polys = {}

        # Display stuff that should be removed from constructor
        poly.xy = calc_display_coords(poly.basecoords, poly.theta)
        poly.lines = poly._make_lines(line_color, line_width)
        poly.handle = poly._make_handle_line()
        poly.species = species
        if ax is not None:
            poly.axes_init(ax)

    def axes_init(poly, ax):
        species = poly.species
        metadata = poly.metadata
        if isinstance(metadata, ut.LazyDict):
            metadata_ = ut.dict_subset(metadata, metadata.cached_keys())
        else:
            metadata_ = metadata
        poly.species_tag = ax.text(
            # tagpos[0], tagpos[1],
            0,
            0,
            species,
            bbox={'facecolor': 'white', 'alpha': 0.8},
            verticalalignment='top',
        )
        poly.metadata_tag = ax.text(
            0,
            0,
            # tagpos[0] + 5, tagpos[1] + 80,
            ut.repr3(metadata_, nobr=True),
            bbox={'facecolor': 'white', 'alpha': 0.7},
            verticalalignment='top',
        )
        # ???
        poly.species_tag.remove()  # eliminate "leftover" copies
        poly.metadata_tag.remove()
        #
        poly.update_display_coords()

    def move_to_back(poly):
        # FIXME: doesnt work exactly
        # Probalby need to do in the context of other polys
        zorder = 0
        poly.set_zorder(zorder)
        poly.lines.set_zorder(zorder)
        poly.handle.set_zorder(zorder)

    def __nice__(poly):
        return '(num=%r)' % (poly.num)

    def add_to_axis(poly, ax):
        ax.add_patch(poly)
        ax.add_line(poly.lines)
        ax.add_line(poly.handle)

    def remove_from_axis(poly, ax):
        poly.remove()
        poly.lines.remove()
        poly.handle.remove()

    def draw_self(poly, ax, show_species_tags=False, editable=True):
        ax.draw_artist(poly)
        if not editable and poly.lines.get_marker():
            poly.lines.set_marker('')
        elif editable and not poly.lines.get_marker():
            poly.lines.set_marker('o')
        ax.draw_artist(poly.lines)
        if editable:
            ax.draw_artist(poly.handle)
        if editable and show_species_tags:
            # Hack to fix matplotlib 1.5 bug
            poly.species_tag.figure = ax.figure
            poly.metadata_tag.figure = ax.figure
            ax.draw_artist(poly.species_tag)
            ax.draw_artist(poly.metadata_tag)

    def _make_lines(poly, line_color, line_width):
        """ verts - list of (x, y) tuples """
        _xs, _ys = list(zip(*poly.xy))
        color = np.array(line_color)
        marker_face_color = line_color
        line_kwargs = {'lw': line_width, 'color': color, 'mfc': marker_face_color}
        lines = mpl.lines.Line2D(
            _xs, _ys, marker='o', alpha=1, animated=True, **line_kwargs
        )
        return lines

    def _make_handle_line(poly):
        _xs, _ys = list(zip(*poly.calc_handle_display_coords()))
        line_width = 4
        line_color = (0, 1, 0)
        color = np.array(line_color)
        marker_face_color = line_color
        line_kwargs = {'lw': line_width, 'color': color, 'mfc': marker_face_color}
        lines = mpl.lines.Line2D(
            _xs, _ys, marker='o', alpha=1, animated=True, **line_kwargs
        )
        return lines

    def calc_tag_position(poly):
        r"""

        CommandLine:
            python -m wbia.plottool.interact_annotations --test-calc_tag_position --show

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.plottool.interact_annotations import *  # NOQA
            >>> poly = ut.DynStruct()
            >>> poly.basecoords = vt.verts_from_bbox([0, 0, 400, 400], True)
            >>> poly.theta = 0
            >>> poly.xy = vt.verts_from_bbox([0, 0, 400, 400], True)
            >>> tagpos = poly.calc_tag_position()
            >>> print('tagpos = %r' % (tagpos,))
        """
        points = [
            [max(list(zip(*poly.basecoords))[0]), min(list(zip(*poly.basecoords))[1])]
        ]
        tagpos = rotate_points_around(points, poly.theta, *points_center(poly.xy))[0]
        return tagpos

    def calc_handle_display_coords(poly):
        img_h = poly.manager.img.shape[0]
        handle_length = img_h // 32
        # MIN_HANDLE_LENGTH = 25
        # handle_length = MIN_HANDLE_LENGTH
        # handle_length = max(MIN_HANDLE_LENGTH, (h / 4))
        cx, cy = points_center(poly.xy)
        w, h = vt.get_pointset_extent_wh(np.array(poly.basecoords))
        x0, y0 = cx, (cy - (h / 2))  # start at top edge
        x1, y1 = (x0, y0 - handle_length)
        pts = [(x0, y0), (x1, y1)]
        pts = rotate_points_around(pts, poly.theta, cx, cy)
        return pts

    def update_color(poly, selected=False, editing_parts=False):
        if editing_parts:
            poly.lines.set_color(df2.PINK)
        elif selected:
            # Add selected color
            sel_color = df2.ORANGE if poly.is_orig else df2.LIGHT_BLUE
            poly.lines.set_color(sel_color)
        else:
            line = poly.lines
            line_color = line.get_color()
            desel_color = df2.WHITE if poly.is_orig else df2.LIGHTGRAY
            if np.any(line_color != np.array(desel_color)):
                line.set_color(np.array(desel_color))

    def update_lines(poly):
        poly.lines.set_data(list(zip(*poly.xy)))
        poly.handle.set_data(list(zip(*poly.calc_handle_display_coords())))

    def set_species(poly, text):
        poly.tctext = text
        poly.species_tag.set_text(text)

    def increment_species(poly, amount=1):
        if len(poly.tab_list) > 0:
            tci = (poly.tcindex + amount) % len(poly.tab_list)
            poly.tcindex = tci
            # All tab is going to do is go through the possibilities
            poly.species_tag.set_text(poly.tab_list[poly.tcindex])

    def resize_poly(poly, x, y, idx, ax):
        """
        Resize a rectangle using idx as the given anchor point. Respects
        current rotation.

        CommandLine:
            python -m wbia.plottool.interact_annotations --exec-resize_poly --show

        Example:
            >>> # DISABLE_DOCTEST
            >>> from wbia.plottool.interact_annotations import *  # NOQA
            >>> (h, w) = img.shape[0:2]
            >>> x1, y1 = 10, 10
            >>> x2, y2 = w - 10,  h - 10
            >>> coords = ((x1, y1), (x1, y2), (x2, y2), (x2, y1))
            >>> x = 3 * w / 4
            >>> y = 3 * h / 4
            >>> idx = 3
            >>> resize_poly(poly, x, y, idx)
            >>> update_UI()
            >>> import wbia.plottool as pt
            >>> pt.show_if_requested()
        """
        # TODO: allow resize by middle click to scale from the center
        # the minus one is because the last coordinate is duplicated (by
        # matplotlib) to get a closed polygon
        tmpcoords = poly.xy[:-1]
        idx = idx % len(tmpcoords)
        previdx = (idx - 1) % len(tmpcoords)
        nextidx = (idx + 1) % len(tmpcoords)
        (dx, dy) = (x - poly.xy[idx][0], y - poly.xy[idx][1])
        # Fudge factor is due to gravity vectors constants
        fudge_factor = (idx) * TAU / 4
        poly_theta = poly.theta + fudge_factor

        polar_idx2prev = polarDelta(tmpcoords[idx], tmpcoords[previdx])
        polar_idx2next = polarDelta(tmpcoords[idx], tmpcoords[nextidx])
        tmpcoords[idx] = (tmpcoords[idx][0] + dx, tmpcoords[idx][1] + dy)
        mag_delta = np.linalg.norm((dx, dy))
        theta_delta = np.arctan2(dy, dx)
        theta_rot = theta_delta - (poly_theta + TAU / 4)
        rotx = mag_delta * np.cos(theta_rot)
        roty = mag_delta * np.sin(theta_rot)
        polar_idx2prev[0] -= rotx
        polar_idx2next[0] += roty
        tmpcoords[previdx] = apply_polarDelta(polar_idx2prev, tmpcoords[idx])
        tmpcoords[nextidx] = apply_polarDelta(polar_idx2next, tmpcoords[idx])

        # rotate the points by -theta to get the "unrotated" points for use as
        # basecoords
        tmpcoords = rotate_points_around(tmpcoords, -poly.theta, *points_center(poly.xy))
        # ensure the poly is closed, matplotlib might do this, but I'm not sure
        # if it preserves the ordering we depend on, even if it does add the
        # point
        tmpcoords = tmpcoords[:] + [tmpcoords[0]]

        dispcoords = calc_display_coords(tmpcoords, poly.theta)

        if check_valid_coords(ax, dispcoords) and check_min_wh(tmpcoords):
            poly.basecoords = tmpcoords
        poly.update_display_coords()

    def rotate_poly(poly, dtheta, ax):
        coords_lis = calc_display_coords(poly.basecoords, poly.theta + dtheta)
        if check_valid_coords(ax, coords_lis):
            poly.theta += dtheta
            poly.update_display_coords()

    def move_poly(poly, dx, dy, ax):
        new_coords = [(x + dx, y + dy) for (x, y) in poly.basecoords]
        coords_list = calc_display_coords(new_coords, poly.theta)
        if check_valid_coords(ax, coords_list):
            poly.basecoords = new_coords
            poly.update_display_coords()

    def update_display_coords(poly):
        poly.xy = calc_display_coords(poly.basecoords, poly.theta)
        tag_pos = poly.calc_tag_position()
        poly.species_tag.set_position((tag_pos[0] + 5, tag_pos[1]))
        poly.metadata_tag.set_position((tag_pos[0] + 5, tag_pos[1] + 50))

    def print_info(poly):
        logger.info('poly = %r' % (poly,))
        logger.info('poly.tag_text = %r' % (poly.species_tag.get_text(),))
        logger.info('poly.metadata = %r' % (poly.metadata,))

    def get_poly_mask(poly, shape):
        h, w = shape[0:2]
        y, x = np.mgrid[:h, :w]
        points = np.transpose((x.ravel(), y.ravel()))
        verts = poly.xy
        path = mpl.path.Path(verts)
        mask = path.contains_points(points)
        # mask = nxutils.points_inside_poly(points, verts)
        return mask.reshape(h, w)

    def is_near_handle(poly, xy_pt, max_dist):
        line = poly.calc_handle_display_coords()
        return is_within_distance_from_line(xy_pt, line, max_dist)

    @property
    def size(poly):
        return vt.bbox_from_verts(poly.xy)[2:4]


@six.add_metaclass(ut.ReloadingMetaclass)
class AnnotationInteraction(abstract_interaction.AbstractInteraction):
    """
    An interactive polygon editor.

    SeeAlso:
        wbia.viz.interact.interact_annotations2
        (ensure that any updates here are propogated there)

    Args:
        verts_list (list) : list of lists of (float, float)
            List of (x, y) coordinates used as vertices of the polygon.
    """

    # --- Initialization and Figure Widgets
    def __init__(
        self,
        img,
        img_ind=None,
        commit_callback=None,
        verts_list=None,
        bbox_list=None,
        theta_list=None,
        species_list=None,
        metadata_list=None,
        line_width=4,
        line_color=(1, 1, 1),
        face_color=(0, 0, 0),
        fnum=None,
        default_species=DEFAULT_SPECIES_TAG,
        next_callback=None,
        prev_callback=None,
        do_mask=False,
        valid_species=[],
        **kwargs,
    ):
        super(AnnotationInteraction, self).__init__(fnum=fnum, **kwargs)

        self.valid_species = valid_species
        self.commit_callback = commit_callback  # commit_callback
        self.but_width = 0.14
        # self.but_height = .08
        self.next_prev_but_height = 0.08
        self.but_height = self.next_prev_but_height - 0.01
        self.callback_funcs = dict(
            [
                ('close_event', self.on_close),
                ('draw_event', self.draw_callback),
                ('button_press_event', self.on_click),
                ('button_release_event', self.on_click_release),
                ('figure_leave_event', self.on_figure_leave),
                ('key_press_event', self.on_key_press),
                ('motion_notify_event', self.on_motion),
                ('pick_event', self.on_pick),
                # ('resize_event', self.on_resize),
            ]
        )
        self.mpl_callback_ids = {}
        self.img = img
        self.show_species_tags = True
        self.max_dist = 10

        def _reinitialize_variables():
            self.do_mask = do_mask
            self.img_ind = img_ind
            self.species_tag = default_species
            self.showverts = True
            self.fc_default = face_color
            self.mouseX = None  # mouse X coordinate
            self.mouseY = None  # mouse Y coordinate
            self.ind_xy = None
            self._autoinc_polynum = it.count(0)  # num polys in image
            self._poly_held = False  # if any poly is active
            self._selected_poly = None  # active polygon
            self.parent_poly = None  # level of parts heirarchy
            self.background = None
            # Ensure nothing is down
            self.reset_mouse_state()

        _reinitialize_variables()
        # hack involving exploting lexical scoping to save defaults for a
        # restore operation
        self.reinitialize_variables = _reinitialize_variables

        try:
            self.fig = df2.figure(fnum=self.fnum, doclf=True, docla=True)
            df2.close_figure(self.fig)
        except AttributeError:
            pass
        self.fig = df2.figure(fnum=self.fnum, doclf=True, docla=True)

        self.reinitialize_figure(fnum=self.fnum)
        assert verts_list is None or bbox_list is None, 'only one can be specified'
        # bbox_list will get converted to verts_list
        if verts_list is not None:
            bbox_list = vt.bboxes_from_vert_list(verts_list)
        if bbox_list is not None:
            verts_list = [vt.verts_from_bbox(bbox) for bbox in bbox_list]
        if theta_list is None:
            theta_list = [0 for _ in verts_list]
        if species_list is None:
            species_list = [self.species_tag for _ in verts_list]
        if metadata_list is None:
            metadata_list = [None for _ in verts_list]

        # Create the list of polygons
        self.handle_polygon_creation(bbox_list, theta_list, species_list, metadata_list)
        self._ind = None  # the active vert
        self._current_rotate_poly = None

        self.mpl_callback_ids = {}
        self.connect_mpl_callbacks(self.fig.canvas)

        self.add_action_buttons()
        self.update_callbacks(next_callback, prev_callback)

    def reinitialize_figure(self, fnum=None):
        self.fig.clear()
        self.fig.clf()
        # self.fig.cla()
        # ut.qflag()
        self.fnum = fnum
        # logger.info(self.fnum)
        ax = df2.gca()
        # self.fig.ax = ax
        self.ax = ax
        df2.remove_patches(self.ax)
        df2.imshow(self.img, fnum=fnum)

        ax.set_clip_on(False)
        ax.set_title(
            (
                '\n'.join(
                    [
                        'Click and drag to select/move/resize/orient an ANNOTATION',
                        # 'Press enter to clear the species tag of the selected ANNOTATION',
                        'Press tab to cycle through annotation species',
                        # 'Type to edit the ANNOTATION species (press tab to autocomplete)'
                    ]
                )
            )
        )

    def add_action_buttons(self):
        self.append_button(
            'Add Annotation\n' + pretty_hotkey_map(ADD_RECTANGLE_HOTKEY),
            rect=[0.18, 0.015, self.but_width, self.but_height],
            callback=self.add_new_poly,
        )
        # self.append_button(
        #     'Add Full Annotation\n' + pretty_hotkey_map(ADD_RECTANGLE_FULL_HOTKEY),
        #     rect=[0.34, 0.015, self.but_width, self.but_height],
        #     callback=ut.partial(self.add_new_poly, full=True)
        # )
        self.append_button(
            'Delete Annotation\n' + pretty_hotkey_map(DEL_RECTANGLE_HOTKEY),
            rect=[0.50, 0.015, self.but_width, self.but_height],
            callback=self.delete_current_poly,
        )
        self.append_button(
            'Save and Exit\n' + pretty_hotkey_map(ACCEPT_SAVE_HOTKEY),
            rect=[0.66, 0.015, self.but_width, self.but_height],
            callback=self.save_and_exit,
        )

    def disconnect_mpl_callbacks(self, canvas):
        """ disconnects all connected matplotlib callbacks """
        for name, callbackid in six.iteritems(self.mpl_callback_ids):
            canvas.mpl_disconnect(callbackid)
        self.mpl_callback_ids = {}

    def connect_mpl_callbacks(self, canvas):
        """disconnects matplotlib callbacks specified in the
        self.mpl_callback_ids dict"""
        # http://matplotlib.org/1.3.1/api/backend_bases_api.html
        # Create callback ids
        self.disconnect_mpl_callbacks(canvas)
        self.mpl_callback_ids = {
            name: canvas.mpl_connect(name, func)
            for name, func in six.iteritems(self.callback_funcs)
        }
        self.fig.canvas = canvas

    # --- Updates

    def update_callbacks(self, next_callback, prev_callback):
        self.prev_callback = prev_callback
        self.next_callback = next_callback
        # Hack because the callbacks actually need to be wrapped
        _next_callback = None if self.next_callback is None else self.next_image
        _prev_callback = None if self.prev_callback is None else self.prev_image
        self.append_button(
            'Previous Image\n' + pretty_hotkey_map(PREV_IMAGE_HOTKEYS),
            rect=[0.02, 0.01, self.but_width, self.next_prev_but_height],
            callback=_prev_callback,
        )
        self.append_button(
            'Next Image\n' + pretty_hotkey_map(NEXT_IMAGE_HOTKEYS),
            rect=[0.82, 0.01, self.but_width, self.next_prev_but_height],
            callback=_next_callback,
        )

    def update_image_and_callbacks(
        self,
        img,
        bbox_list,
        theta_list,
        species_list,
        metadata_list,
        next_callback,
        prev_callback,
    ):
        self.disconnect_mpl_callbacks(self.fig.canvas)
        for poly in six.itervalues(self.polys):
            poly.remove()
        self.polys = {}
        self.reinitialize_variables()
        self.img = img
        self.reinitialize_figure(fnum=self.fnum)
        self.handle_polygon_creation(bbox_list, theta_list, species_list, metadata_list)
        self.add_action_buttons()
        self.draw()
        self.connect_mpl_callbacks(self.fig.canvas)
        self.update_callbacks(next_callback, prev_callback)
        logger.info('[interact_annot] drawing')
        self.draw()
        self.update_UI()

    def _update_poly_colors(self):
        for poly in six.itervalues(self.uneditable_polys):
            poly.update_color()
        for ind, poly in six.iteritems(self.editable_polys):
            assert poly.num == ind
            selected = poly is self._selected_poly
            editing_parts = poly is self.parent_poly
            poly.update_color(selected, editing_parts)
        self.draw()

    def _update_poly_lines(self):
        for poly in six.itervalues(self.uneditable_polys):
            # self.last_vert_ind = len(poly.xy) - 1
            poly.update_lines()
        for poly in six.itervalues(self.editable_polys):
            self.last_vert_ind = len(poly.xy) - 1
            poly.update_lines()

    def update_UI(self):
        self._update_poly_lines()
        self._update_poly_colors()
        self.fig.canvas.restore_region(self.background)
        self.draw_artists()
        self.fig.canvas.blit(self.ax.bbox)

    def draw_artists(self):
        for poly in six.itervalues(self.uneditable_polys):
            poly.draw_self(self.ax, editable=False)
        for poly in six.itervalues(self.editable_polys):
            poly.draw_self(self.ax, self.show_species_tags)

    # --- Data Matainence / Other

    @property
    def uneditable_polys(self):
        if self.in_edit_parts_mode:
            return {self.parent_poly.num: self.parent_poly}
            # return self.polys
        else:
            return {}

    @property
    def editable_polys(self):
        # return self.polys
        if self.in_edit_parts_mode:
            return self.parent_poly.child_polys
        else:
            if self.polys is None:
                self.polys = {}
            return self.polys

    def get_poly_under_cursor(self, x, y):
        """
        get the index of the vertex under cursor if within max_dist tolerance
        """
        # Remove any deleted polygons
        poly_dict = {k: v for k, v in self.editable_polys.items() if v is not None}
        if len(poly_dict) > 0:
            poly_inds = list(poly_dict.keys())
            poly_list = ut.take(poly_dict, poly_inds)
            # Put polygon coords into figure space
            poly_pts = [
                poly.get_transform().transform(np.asarray(poly.xy)) for poly in poly_list
            ]
            # Find the nearest vertex from the annotations
            ind_dist_list = [vt.nearest_point(x, y, polypts) for polypts in poly_pts]
            dist_lists = ut.take_column(ind_dist_list, 1)
            min_idx = np.argmin(dist_lists)
            sel_polyind = poly_inds[min_idx]
            sel_vertx, sel_dist = ind_dist_list[min_idx]
            # Ensure nearest distance is within threshold
            if sel_dist >= self.max_dist ** 2:
                sel_polyind, sel_vertx = (None, None)
        else:
            sel_polyind, sel_vertx = (None, None)
        return sel_polyind, sel_vertx

    def get_most_recently_added_poly(self):
        if len(self.editable_polys) == 0:
            return None
        else:
            # most recently added polygon has the highest index
            poly_ind = max(list(self.editable_polys.keys()))
            return self.editable_polys[poly_ind]

    def new_polygon(
        self,
        verts,
        theta,
        species,
        fc=(0, 0, 0),
        line_color=(1, 1, 1),
        line_width=4,
        is_orig=False,
        metadata=None,
    ):
        """ verts - list of (x, y) tuples """
        # create new polygon from verts
        num = six.next(self._autoinc_polynum)
        poly = AnnotPoly(
            ax=self.ax,
            num=num,
            verts=verts,
            theta=theta,
            species=species,
            fc=fc,
            line_color=line_color,
            line_width=line_width,
            is_orig=is_orig,
            metadata=metadata,
            valid_species=self.valid_species,
            manager=self,
        )
        poly.set_picker(self.is_poly_pickable)
        return poly

    def handle_polygon_creation(self, bbox_list, theta_list, species_list, metadata_list):
        """ Maintain original input """
        assert bbox_list is not None
        if theta_list is None:
            theta_list = [0.0 for _ in range(len(bbox_list))]
        if species_list is None:
            species_list = ['' for _ in range(len(bbox_list))]
        assert len(bbox_list) == len(theta_list), 'inconconsitent data1'
        assert len(bbox_list) == len(species_list), 'inconconsitent data2'
        assert len(bbox_list) == len(metadata_list), 'inconconsitent data2'
        self.original_indices = list(range(len(bbox_list)))
        self.original_bbox_list = bbox_list
        self.original_theta_list = theta_list
        self.original_species_list = species_list
        self.original_metadata_list = metadata_list
        # Convert bbox to verticies
        verts_list = [vt.verts_from_bbox(bbox) for bbox in bbox_list]
        for verts in verts_list:
            verts = np.array(verts)
            for vert in verts:
                enforce_dims(self.ax, vert)
        # Create polygons
        poly_list = [
            self.new_polygon(verts_, theta, species, is_orig=True, metadata=metadata)
            for (verts_, theta, species, metadata) in zip(
                verts_list, theta_list, species_list, metadata_list
            )
        ]
        self.polys = {poly.num: poly for poly in poly_list}
        if len(self.polys) != 0:
            # Select poly with largest area
            wh_list = np.array([poly.size for poly in six.itervalues(self.polys)])
            poly_index = list(self.polys.keys())[wh_list.prod(axis=1).argmax()]
            self._selected_poly = self.polys[poly_index]
            self._update_poly_colors()
            self._update_poly_lines()
        else:
            self._selected_poly = None
        # Add polygons to the axis
        for poly in six.itervalues(self.polys):
            poly.add_to_axis(self.ax)
        # Give polygons mpl change callbacks
        # for poly in six.itervalues(self.polys):
        #    poly.add_callback(self.poly_changed)

    # --- Actions

    def add_new_poly(self, event=None, full=False):
        """ Adds a new annotation to the image """
        if full:
            (h, w) = self.img.shape[0:2]
            x1, y1 = 1, 1
            x2, y2 = w - 1, h - 1
            coords = ((x1, y1), (x1, y2), (x2, y2), (x2, y1))
        else:
            if self._selected_poly is not None:
                defaultshape_polys = {self._selected_poly.num: self._selected_poly}
            else:
                defaultshape_polys = self.editable_polys
            coords = default_vertices(
                self.img, defaultshape_polys, self.mouseX, self.mouseY
            )

        poly = self.new_polygon(verts=coords, theta=0, species=self.species_tag)
        poly.parent = self.parent_poly

        # Add to the correct place in current heirarchy
        self.editable_polys[poly.num] = poly
        poly.add_to_axis(self.ax)

        # self.polys[poly.num] = poly

        # poly.add_callback(self.poly_changed)
        self._ind = None  # the active vert
        self._selected_poly = self.get_most_recently_added_poly()
        self._update_poly_lines()
        self._update_poly_colors()
        self.draw()

    def delete_current_poly(self, event=None):
        """
        Removes an annotation
        """
        if self._selected_poly is None:
            logger.info('[interact_annot] No polygon selected to delete')
        else:
            logger.info('[interact_annot] delete annot')
            poly = self._selected_poly
            # self.polys.pop(poly.num)
            del self.editable_polys[poly.num]
            # remove the poly from the figure itself
            poly.remove_from_axis(self.ax)
            # reset anything that has to do with current poly
            self._selected_poly = self.get_most_recently_added_poly()
            self._poly_held = False
            if self._selected_poly is not None:
                self._update_poly_colors()
            self.draw()

    def edit_poly_parts(self, poly):
        if poly is None and self.parent_poly is not None:
            self._selected_poly = self.parent_poly
        logger.info('self.parent_poly = %r' % (self.parent_poly,))
        self.parent_poly = poly
        if poly is not None:
            self._selected_poly = self.get_most_recently_added_poly()
        logger.info('self._selected_poly = %r' % (self._selected_poly,))
        if poly is None:
            self.ax.imshow(vt.convert_colorspace(self.img, 'RGB'))
        else:
            # Mask the part of the image not belonging to the annotation
            mask = poly.get_poly_mask(self.img.shape)
            masked_img = apply_mask(self.img, mask)
            self.ax.imshow(vt.convert_colorspace(masked_img, 'RGB'))
        self._update_poly_colors()

    @property
    def in_edit_parts_mode(self):
        return self.parent_poly is not None

    def toggle_species_label(self):
        logger.info('[interact_annot] toggle_species_label()')
        self.show_species_tags = not self.show_species_tags
        self.update_UI()

    def save_and_exit(self, event, do_close=True):
        """
        The Save and Exit Button

        write a callback to redraw viz for bbox_list
        """
        logger.info('[interact_annot] Pressed Accept Button')

        def _get_annottup_list():
            annottup_list = []
            indices_list = []
            # theta_list = []
            for poly in six.itervalues(self.polys):
                assert poly is not None
                index = poly.num
                bbox = tuple(map(int, vt.bbox_from_verts(poly.basecoords)))
                theta = poly.theta
                species = poly.species_tag.get_text()
                annottup = (bbox, theta, species)
                indices_list.append(index)
                annottup_list.append(annottup)
            return indices_list, annottup_list

        def _send_back_annotations():
            logger.info('[interact_annot] _send_back_annotations')
            indices_list, annottup_list = _get_annottup_list()
            # Delete if index is in original_indices but no in indices_list
            deleted_indices = list(set(self.original_indices) - set(indices_list))
            changed_indices = []
            unchanged_indices = []  # sanity check
            changed_annottups = []
            new_annottups = []
            original_annottup_list = list(
                zip(
                    self.original_bbox_list,
                    self.original_theta_list,
                    self.original_species_list,
                )
            )
            for index, annottup in zip(indices_list, annottup_list):
                # If the index is not in the originals then it is new
                if index not in self.original_indices:
                    new_annottups.append(annottup)
                else:
                    if annottup not in original_annottup_list:
                        changed_annottups.append(annottup)
                        changed_indices.append(index)
                    else:
                        unchanged_indices.append(index)
            self.commit_callback(
                unchanged_indices,
                deleted_indices,
                changed_indices,
                changed_annottups,
                new_annottups,
            )

        if self.commit_callback is not None:
            _send_back_annotations()
        # Make mask from selection
        if self.do_mask is True:
            self.fig.clf()
            self.ax = ax = self.fig.subplot(111)
            mask_list = [
                poly.get_poly_mask(self.img.shape) for poly in six.itervalues(self.polys)
            ]
            if len(mask_list) == 0:
                logger.info('[interact_annot] No polygons to make mask out of')
                return 0
            mask = mask_list[0]
            for mask_ in mask_list:
                mask = np.maximum(mask, mask_)
            # mask = self.get_poly_mask()
            # User must close previous figure
            # Modify the image with the mask
            masked_img = apply_mask(self.img, mask)
            # show the modified image
            ax.imshow(masked_img)
            ax.title('Region outside of mask is darkened')

            ax.figure.show()
            return

        logger.info('[interact_annot] Accept Over')
        if do_close:
            df2.close_figure(self.fig)

    # --- Connected Slots and Callbacks

    def next_image(self, event):
        if self.next_callback is not None:
            self.next_callback()

    def prev_image(self, event):
        if self.prev_callback is not None:
            self.prev_callback()

    def start(self):
        # FIXME: conform to abstract_interaction start conventions
        # self._ensure_running()
        # self.show_page()
        self.show()

    def show(self):
        self.draw()
        self.bring_to_front()

    def draw_callback(self, event):
        self.background = self.fig.canvas.copy_from_bbox(self.ax.bbox)
        self.draw_artists()

    def _show_poly_context_menu(self, event):
        def _make_options():
            metadata = self._selected_poly.metadata
            options = []
            options += [
                # ('Foo: ',  ut.partial(print, 'bar')),
                # ('Move to back ',  self._selected_poly.move_to_back),
                ('PolyInfo: ', self._selected_poly.print_info),
            ]
            if isinstance(metadata, ut.LazyDict):
                options += metadata.nocache_eval('annot_context_options')
            return options

        options = _make_options()
        self.show_popup_menu(options, event)

    def is_poly_pickable(self, artist, event):
        if artist.num in self.editable_polys:
            mouse_xy = event.x, event.y
            hit = artist.contains_point(mouse_xy)
        else:
            hit = False
        # import utool
        # utool.embed()
        props = {'dblclick': event.dblclick}
        return hit, props

    def on_pick(self, event):
        """ Makes selected polygon translucent """
        if self.debug > 0 or True:
            logger.info('[interact_annot] on_pick')
        if not self._poly_held:
            artist = event.artist
            logger.info('[interact_annot] picked artist = %r' % (artist,))
            self._selected_poly = artist
            self._poly_held = True
            if event.dblclick and not self.in_edit_parts_mode:
                self.edit_poly_parts(self._selected_poly)
                pass
        # x, y = event.mouseevent.xdata, event.mouseevent.xdata

    def on_click(self, event):
        """
        python -m wbia.viz.interact.interact_annotations2 --test-ishow_image2 --show
        """
        super(AnnotationInteraction, self).on_click(event)

        if self._ind is not None:
            self._ind = None
            return
        if not self.showverts:
            return
        if event.inaxes is None:
            return

        if len(self.editable_polys) == 0:
            logger.info('[interact_annot] No polygons on screen')
            return

        # Right click - context menu
        if event.button == self.RIGHT_BUTTON:
            self._show_poly_context_menu(event)
        # Left click, indicate that a mouse button is down
        if event.button == self.LEFT_BUTTON:
            # if event.dblclick and not self.in_edit_parts_mode:
            #    # On double click enter a single annotation to annotation parts
            #    #logger.info("DOUBLECLICK")
            #    #self.edit_poly_parts(self._selected_poly)
            if event.key == 'shift':
                self._current_rotate_poly = self._selected_poly
            else:
                # Determine if we are clicking the rotation line
                mouse_xy = (event.xdata, event.ydata)
                for poly in six.itervalues(self.editable_polys):
                    if poly.is_near_handle(mouse_xy, self.max_dist):
                        self._current_rotate_poly = poly
                        break
                if event.dblclick:
                    # Reset rotation
                    if self._current_rotate_poly is not None:
                        self._current_rotate_poly.theta = 0
                        self._current_rotate_poly.update_display_coords()

        polyind, self._ind = self.get_poly_under_cursor(event.x, event.y)

        if self._ind is not None and polyind is not None:
            self._selected_poly = self.editable_polys[polyind]
            if self._selected_poly is None:
                return
            self.ind_xy = self._selected_poly.xy[self._ind]
            self._poly_held = True
            self._selected_poly.anchor_idx = self._ind

        self.mouseX, self.mouseY = event.xdata, event.ydata

        if self._poly_held is True or self._ind is not None:
            self._selected_poly.set_alpha(0.2)
            self._update_poly_colors()

        self._update_poly_colors()
        self._update_poly_lines()

        if self.background is not None:
            self.fig.canvas.restore_region(self.background)
        else:
            logger.info(
                '[interact_annot] error: self.background is none.' ' Trying refresh.'
            )
            self.fig.canvas.restore_region(self.background)
            self.background = self.fig.canvas.copy_from_bbox(self.ax.bbox)

        # Redraw blitted objects
        self.draw_artists()
        self.fig.canvas.blit(self.ax.bbox)

    def on_motion(self, event):
        if ut.VERBOSE:
            logger.info('[interact_annot] on_motion')
            logger.info('[interact_annot] Got key: %r' % event.key)
        super(AnnotationInteraction, self).on_motion(event)
        # uses boolean punning for terseness
        lastX = self.mouseX or None
        lastY = self.mouseY or None
        # Allow for getting coordinates outside the axes
        ax = self.ax
        mousePos = [event.x, event.y]
        self.mouseX, self.mouseY = ax.transData.inverted().transform(mousePos)
        deltaX = lastX is not None and self.mouseX - lastX
        deltaY = lastY is not None and self.mouseY - lastY

        if not self.showverts:
            return

        # if self.in_edit_parts_mode:
        #    return

        quick_resize = self._poly_held is True and (
            (event.button == self.MIDDLE_BUTTON)
            or (event.button == self.RIGHT_BUTTON)
            or (event.button == self.LEFT_BUTTON and event.key == 'ctrl')
        )

        if self._poly_held is True and self._ind is not None:
            # Resize by dragging corner
            self._selected_poly.resize_poly(self.mouseX, self.mouseY, self._ind, self.ax)
            self._selected_poly.anchor_idx = self._ind
        elif quick_resize:
            # Quick resize with special click
            anchor_idx = self._selected_poly.anchor_idx
            idx = (anchor_idx + 2) % 4  # choose opposite anchor point
            self._selected_poly.resize_poly(self.mouseX, self.mouseY, idx, self.ax)
        elif self._current_rotate_poly:
            # Rotate using handle
            cx, cy = points_center(self._current_rotate_poly.xy)
            theta = np.arctan2(cy - self.mouseY, cx - self.mouseX) - TAU / 4
            dtheta = theta - self._current_rotate_poly.theta
            self._current_rotate_poly.rotate_poly(dtheta, self.ax)
        elif self._ind is None and event.button == self.LEFT_BUTTON:
            # Translate by dragging inside annot
            flag = deltaX is not None and deltaY is not None
            if self._poly_held is True and flag:
                self._selected_poly.move_poly(deltaX, deltaY, self.ax)
            self._ind = None
        else:
            return
        self.update_UI()

    def on_click_release(self, event):
        super(AnnotationInteraction, self).on_click_release(event)

        # if self._poly_held is True:
        self._poly_held = False

        self._current_rotate_poly = None

        if not self.showverts:
            return

        if self._selected_poly is None:
            return

        _flag = (
            self._ind is None
            or self._poly_held is False
            or (
                self._ind is not None
                and self.is_down['left'] is True
                and self._selected_poly is not None
            )
        )
        if _flag:
            self._selected_poly.set_alpha(0)
            # self._selected_poly.set_facecolor('white')

        self.update_UI()

        if self._ind is None:
            return

        if len(self.editable_polys) == 0:
            logger.info('[interact_annot] No polygons on screen')
            return

        if self._selected_poly is None:
            logger.info(
                '[interact_annot] WARNING: Polygon unknown.' ' Using default. (2)'
            )
            self._selected_poly = self.get_most_recently_added_poly()

        curr_xy = self._selected_poly.xy[self._ind]

        if self.ind_xy is not None:
            if np.all(np.fabs(self.ind_xy - curr_xy) < 3):
                return

        self._ind = None
        self._poly_held = False

        self.draw()

    def on_figure_leave(self, event):
        if self.debug > 0:
            logger.info('[interact_annot] figure leave')
        # self.print_status()
        # self.on_click_release(event)
        self._poly_held = False
        self._ind = None
        self.reset_mouse_state()
        # self.print_status()

    def on_key_press(self, event):
        if self.debug > 0:
            logger.info('[interact_annot] on_key_press')
            logger.info('[interact_annot] Got key: %r' % event.key)
        logger.info('[interact_annot] Got key: %r' % event.key)
        if not event.inaxes:
            return

        if event.key == ACCEPT_SAVE_HOTKEY:
            self.save_and_exit(event)
        elif event.key == ADD_RECTANGLE_HOTKEY:
            self.add_new_poly()
        elif event.key == ADD_RECTANGLE_FULL_HOTKEY:
            self.add_new_poly(full=True)
        elif event.key == DEL_RECTANGLE_HOTKEY:
            self.delete_current_poly()
        elif event.key == TOGGLE_LABEL_HOTKEY:
            self.toggle_species_label()

        if re.match('escape', event.key):
            self.edit_poly_parts(None)

        if re.match('^backspace$', event.key):
            self._selected_poly.set_species(DEFAULT_SPECIES_TAG)
        if re.match('^tab$', event.key):
            self._selected_poly.increment_species(amount=1)
        if re.match(r'^ctrl\+tab$', event.key):
            self._selected_poly.increment_species(amount=-1)

        # NEXT ANND PREV COMMAND
        def _matches_hotkey(key, hotkeys):
            return any(
                [re.match(hk, key) is not None for hk in ut.ensure_iterable(hotkeys)]
            )

        if _matches_hotkey(event.key, PREV_IMAGE_HOTKEYS):
            self.prev_image(event)
        if _matches_hotkey(event.key, NEXT_IMAGE_HOTKEYS):
            self.next_image(event)
        self.draw()

    # def poly_changed(self, poly):
    #    """ this method is called whenever the polygon object is called """
    #    logger.info('poly_changed poly=%r' % (poly,))
    #    # only copy the artist props to the line (except visibility)
    #    #vis = poly.lines.get_visible()
    #    #vis = poly.handle.get_visible()
    #    #poly.lines.set_visible(vis)
    #    #poly.handle.set_visible(vis)


def pretty_hotkey_map(hotkeys):
    if hotkeys is None:
        return ''
    hotkeys = [hotkeys] if not isinstance(hotkeys, list) else hotkeys
    mapping = {
        # 'right': 'right arrow',
        # 'left':  'left arrow',
    }
    mapped_hotkeys = [mapping.get(hk, hk) for hk in hotkeys]
    hotkey_str = '(' + ut.conj_phrase(mapped_hotkeys, 'or') + ')'
    return hotkey_str


def apply_mask(img, mask):
    masked_img = img.copy()
    masked_img[~mask] = np.uint8(np.clip(masked_img[~mask] - 100.0, 0, 255))
    return masked_img


def points_center(pts):
    # the polygons have the first point listed twice in order for them to be
    # drawn as closed, but that point shouldn't be counted twice for computing
    # the center (hence the [:-1] slice)
    return np.array(pts[:-1]).mean(axis=0)


def rotate_points_around(points, theta, ax, ay):
    """
    References:
        http://www.euclideanspace.com/maths/geometry/affine/aroundPoint/matrix2d/
    """
    # TODO: Can use vtool for this
    sin, cos, array = np.sin, np.cos, np.array
    augpts = array([array((x, y, 1)) for (x, y) in points])
    ct = cos(theta)
    st = sin(theta)
    # correct matrix obtained from
    rot_mat = array(
        [(ct, -st, ax - ct * ax + st * ay), (st, ct, ay - st * ax - ct * ay), (0, 0, 1)]
    )
    return [(x, y) for (x, y, z) in rot_mat.dot(augpts.T).T]


def calc_display_coords(oldcoords, theta):
    return rotate_points_around(oldcoords, theta, *points_center(oldcoords))


def polarDelta(p1, p2):
    mag = vt.L2(p1, p2)
    theta = np.arctan2(p2[1] - p1[1], p2[0] - p1[0])
    return [mag, theta]


def apply_polarDelta(poldelt, cart):
    newx = cart[0] + (poldelt[0] * np.cos(poldelt[1]))
    newy = cart[1] + (poldelt[0] * np.sin(poldelt[1]))
    return (newx, newy)


def is_within_distance_from_line(pt, line, max_dist):
    pt = np.array(pt)
    line = np.array(line)
    return vt.distance_to_lineseg(pt, line[0], line[1]) <= max_dist


def check_min_wh(coords):
    """
    Depends on hardcoded indices, which is inelegant, but
    we're already depending on those for the FUDGE_FACTORS
    array above
    0----1
    |    |
    3----2
    """
    MIN_W = 5
    MIN_H = 5
    # the seperate 1 and 2 variables are not strictly necessary, but
    # provide a sanity check to ensure that we're dealing with the
    # right shape
    # w, h = vt.get_pointset_extent_wh(np.array(coords))
    w1 = coords[1][0] - coords[0][0]
    w2 = coords[2][0] - coords[3][0]
    h1 = coords[3][1] - coords[0][1]
    h2 = coords[2][1] - coords[1][1]
    assert np.isclose(w1, w2), 'w1: %r, w2: %r' % (w1, w2)
    assert np.isclose(h1, h2), 'h1: %r, h2: %r' % (h1, h2)
    w, h = w1, h1
    # logger.info('w, h = (%r, %r)' % (w1, h1))
    return (MIN_W < w) and (MIN_H < h)


def default_vertices(img, polys=None, mouseX=None, mouseY=None):
    """Default to rectangle that has a quarter-width/height border."""
    (h, w) = img.shape[0:2]
    # Center the new verts around wherever the mouse is
    if mouseX is not None and mouseY is not None:
        center_x = mouseX
        center_h = mouseY
    else:
        center_x = w // 2
        center_h = h // 2

    if polys is not None and len(polys) > 0:
        # Use the largest polygon size as the default verts
        wh_list = np.array(
            [vt.bbox_from_verts(poly.xy)[2:4] for poly in six.itervalues(polys)]
        )
        w_, h_ = wh_list.max(axis=0) // 2
    else:
        # If no poly exists use 1/4 of the image size
        w_, h_ = (w // 4, h // 4)
    # Get the x/y extents by offseting the centers
    x1, x2 = np.array([center_x, center_x]) + (w_ * np.array([-1, 1]))
    y1, y2 = np.array([center_h, center_h]) + (h_ * np.array([-1, 1]))
    # Clip to bounds
    x1 = max(x1, 1)
    y1 = max(y1, 1)
    x2 = min(x2, w - 1)
    y2 = min(y2, h - 1)
    return ((x1, y1), (x1, y2), (x2, y2), (x2, y1))


def check_valid_coords(ax, coords_list):
    return all([check_dims(ax, xy_pt) for xy_pt in coords_list])


def check_dims(ax, xy_pt, margin=0.5):
    """
    checks if bounding box dims are ok

    Allow the bounding box to go off the image
    so orientations can be done correctly
    """
    num_out = 0
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    if xy_pt[0] < xlim[0] + margin:
        num_out += 1
    if xy_pt[0] > xlim[1] - margin:
        num_out += 1
    if xy_pt[1] < ylim[1] + margin:
        num_out += 1
    if xy_pt[1] > ylim[0] - margin:
        num_out += 1
    return num_out <= 3


def enforce_dims(ax, xy_pt, margin=0.5):
    """
    ONLY USE THIS ON UNROTATED RECTANGLES, as to do otherwise may yield
    arbitrary polygons
    """
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    if xy_pt[0] < xlim[0] + margin:
        xy_pt[0] = xlim[0] + margin
    if xy_pt[0] > xlim[1] - margin:
        xy_pt[0] = xlim[1] - margin
    if xy_pt[1] < ylim[1] + margin:
        xy_pt[1] = ylim[1] + margin
    if xy_pt[1] > ylim[0] - margin:
        xy_pt[1] = ylim[0] - margin
    return True


def test_interact_annots():
    r"""
    CommandLine:
        python -m wbia.plottool.interact_annotations --test-test_interact_annots --show

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.plottool.interact_annotations import *  # NOQA
        >>> import wbia.plottool as pt
        >>> # build test data
        >>> # execute function
        >>> self = test_interact_annots()
        >>> # verify results
        >>> print(self)
        >>> pt.show_if_requested()
    """
    logger.info('[interact_annot] *** START DEMO ***')
    verts_list = [
        ((0, 400), (400, 400), (400, 0), (0, 0), (0, 400)),
        ((400, 700), (700, 700), (700, 400), (400, 400), (400, 700)),
    ]
    # if img is None:
    try:
        img_url = 'http://i.imgur.com/Vq9CLok.jpg'
        img_fpath = ut.grab_file_url(img_url)
        img = vt.imread(img_fpath)
    except Exception as ex:
        logger.info('[interact_annot] cant read zebra: %r' % ex)
        img = np.random.uniform(0, 255, size=(100, 100))
    valid_species = ['species1', 'species2']
    metadata_list = [{'name': 'foo'}, None]
    self = AnnotationInteraction(
        img,
        verts_list=verts_list,
        valid_species=valid_species,
        metadata_list=metadata_list,
        fnum=0,
    )  # NOQA
    return self



# -*- coding: utf-8 -*-
"""
AUTOGENERATED ON 10:14:54 2019/07/09
AutogenCommandLine:
    python -m wbia.control.DB_SCHEMA --test-autogen_db_schema --force-incremental-db-update --write
    python -m wbia.control.DB_SCHEMA --test-autogen_db_schema --force-incremental-db-update --diff=1
    python -m wbia.control.DB_SCHEMA --test-autogen_db_schema --force-incremental-db-update
"""
# -*- coding: utf-8 -*-

# import utool as ut
# (print, rrr, profile) = ut.inject2(__name__)
import logging

logger = logging.getLogger('wbia')

# =======================
# Schema Version Current
# =======================


VERSION_CURRENT = '2.0.0'


def update_current(db, ibs=None):
    db.add_table(
        'annotation_lblannot_relationship',
        [
            ('alr_rowid', 'INTEGER PRIMARY KEY'),
            ('annot_rowid', 'INTEGER NOT NULL'),
            ('lblannot_rowid', 'INTEGER NOT NULL'),
            ('alr_confidence', 'REAL DEFAULT 0.0'),
        ],
        docstr="""
        Used to store one-to-many the relationship between annotations
        (annots) and labels
        """,
        superkeys=[('annot_rowid', 'lblannot_rowid')],
        relates=('annotations', 'lblannot'),
        shortname='alr',
    )

    db.add_table(
        'annotations',
        [
            ('annot_rowid', 'INTEGER PRIMARY KEY'),
            ('annot_parent_rowid', 'INTEGER'),
            ('annot_uuid', 'UUID NOT NULL'),
            ('image_rowid', 'INTEGER NOT NULL'),
            ('annot_xtl', 'INTEGER NOT NULL'),
            ('annot_ytl', 'INTEGER NOT NULL'),
            ('annot_width', 'INTEGER NOT NULL'),
            ('annot_height', 'INTEGER NOT NULL'),
            ('annot_theta', 'REAL DEFAULT 0.0'),
            ('annot_num_verts', 'INTEGER NOT NULL'),
            ('annot_verts', 'TEXT'),
            ('annot_yaw', 'REAL'),
            ('annot_viewpoint', 'TEXT'),
            ('annot_detect_confidence', 'REAL DEFAULT -1.0'),
            ('annot_toggle_reviewed', 'INTEGER DEFAULT 0'),
            ('annot_toggle_multiple', 'INTEGER DEFAULT NULL'),
            ('annot_toggle_interest', 'INTEGER DEFAULT NULL'),
            ('annot_toggle_canonical', 'INTEGER DEFAULT NULL'),
            ('annot_exemplar_flag', 'INTEGER DEFAULT 0'),
            ('annot_note', 'TEXT'),
            ('annot_visual_uuid', 'UUID NOT NULL'),
            ('annot_semantic_uuid', 'UUID'),
            ('name_rowid', 'INTEGER DEFAULT 0'),
            ('species_rowid', 'INTEGER DEFAULT 0'),
            ('annot_quality', 'INTEGER'),
            ('contributor_rowid', 'INTEGER'),
            ('annot_age_months_est_min', 'INTEGER DEFAULT -1'),
            ('annot_age_months_est_max', 'INTEGER DEFAULT -1'),
            ('annot_tag_text', 'TEXT'),
            ('annot_metadata_json', 'TEXT'),
            ('annot_static_encounter', 'TEXT'),
            ('annot_viewpoint_int', 'INTEGER'),
            ('annot_staged_flag', 'INTEGER DEFAULT 0'),
            ('annot_staged_uuid', 'UUID'),
            ('annot_staged_user_identity', 'TEXT'),
            ('annot_staged_metadata_json', 'TEXT'),
        ],
        docstr="""
        Mainly used to store the geometry of the annotation within its parent
        image The one-to-many relationship between images and annotations is
        encoded here
        """,
        superkeys=[('annot_uuid',), ('annot_visual_uuid', 'annot_staged_uuid')],
        shortname='annot',
        extern_tables=['names', 'species', 'images'],
        primary_superkey=('annot_uuid',),
        dependsmap={
            'annot_parent_rowid': (
                'annotations',
                ('annot_rowid',),
                ('annot_visual_uuid',),
            ),
            'contributor_rowid': ('contributors', None, None),
            'image_rowid': ('images', ('image_rowid',), ('image_uuid',)),
            'name_rowid': ('names', ('name_rowid',), ('name_text',)),
            'species_rowid': ('species', ('species_rowid',), ('species_text',)),
        },
    )

    db.add_table(
        'annotgroup_annotation_relationship',
        [
            ('gar_rowid', 'INTEGER PRIMARY KEY'),
            ('annotgroup_rowid', 'INTEGER NOT NULL'),
            ('annot_rowid', 'INTEGER'),
        ],
        docstr="""
        Relationship between annotgroups and annots (many to many mapping) the
        many-to-many relationship between annots and annotgroups is encoded
        here annotgroup_annotation_relationship stands for annotgroup-
        annotation-pairs.
        """,
        superkeys=[('annotgroup_rowid', 'annot_rowid')],
    )

    db.add_table(
        'annotgroups',
        [
            ('annotgroup_rowid', 'INTEGER PRIMARY KEY'),
            ('annotgroup_uuid', 'UUID NOT NULL'),
            ('annotgroup_text', 'TEXT NOT NULL'),
            ('annotgroup_note', 'TEXT NOT NULL'),
        ],
        docstr="""
        List of all annotation groups (annotgroups)
        """,
        superkeys=[('annotgroup_text',)],
    )

    db.add_table(
        'annotmatch',
        [
            ('annotmatch_rowid', 'INTEGER PRIMARY KEY'),
            ('annot_rowid1', 'INTEGER NOT NULL'),
            ('annot_rowid2', 'INTEGER NOT NULL'),
            ('annotmatch_evidence_decision', 'INTEGER'),
            ('annotmatch_confidence', 'INTEGER'),
            ('annotmatch_tag_text', 'TEXT'),
            ('annotmatch_reviewer', 'TEXT'),
            ('annotmatch_posixtime_modified', 'INTEGER'),
            ('annotmatch_count', 'INTEGER'),
            ('annotmatch_meta_decision', 'INTEGER'),
        ],
        docstr="""
        Sparsely stores explicit matching / not matching information. This
        serves as marking weather or not an annotation pair has been reviewed.
        """,
        superkeys=[('annot_rowid1', 'annot_rowid2')],
        relates=('annotations', 'annotations'),
        dependsmap={
            'annot_rowid1': ('annotations', ('annot_rowid',), ('annot_visual_uuid',)),
            'annot_rowid2': ('annotations', ('annot_rowid',), ('annot_visual_uuid',)),
        },
    )

    db.add_table(
        'contributors',
        [
            ('contributor_rowid', 'INTEGER PRIMARY KEY'),
            ('contributor_uuid', 'UUID NOT NULL'),
            ('contributor_tag', 'TEXT'),
            ('contributor_name_first', 'TEXT'),
            ('contributor_name_last', 'TEXT'),
            ('contributor_location_city', 'TEXT'),
            ('contributor_location_state', 'TEXT'),
            ('contributor_location_country', 'TEXT'),
            ('contributor_location_zip', 'TEXT'),
            ('contributor_note', 'TEXT'),
        ],
        docstr="""
        Used to store the contributors to the project
        """,
        superkeys=[('contributor_tag',)],
    )

    db.add_table(
        'image_lblimage_relationship',
        [
            ('glr_rowid', 'INTEGER PRIMARY KEY'),
            ('image_rowid', 'INTEGER NOT NULL'),
            ('lblimage_rowid', 'INTEGER NOT NULL'),
            ('glr_confidence', 'REAL DEFAULT 0.0'),
        ],
        docstr="""
        Used to store one-to-many the relationship between images and labels
        """,
        superkeys=[('image_rowid', 'lblimage_rowid')],
        relates=('images', 'lblimage'),
        shortname='glr',
    )

    db.add_table(
        'images',
        [
            ('image_rowid', 'INTEGER PRIMARY KEY'),
            ('contributor_rowid', 'INTEGER'),
            ('image_uuid', 'UUID NOT NULL'),
            ('image_uri', 'TEXT NOT NULL'),
            ('image_uri_original', 'TEXT NOT NULL'),
            ('image_ext', 'TEXT NOT NULL'),
            ('image_original_name', 'TEXT NOT NULL'),
            ('image_width', 'INTEGER DEFAULT -1'),
            ('image_height', 'INTEGER DEFAULT -1'),
            ('image_time_posix', 'INTEGER DEFAULT -1'),
            ('image_gps_lat', 'REAL DEFAULT -1.0'),
            ('image_gps_lon', 'REAL DEFAULT -1.0'),
            ('image_orientation', 'INTEGER DEFAULT 0'),
            ('image_toggle_enabled', 'INTEGER DEFAULT 0'),
            ('image_toggle_reviewed', 'INTEGER DEFAULT 0'),
            ('image_toggle_cameratrap', 'INTEGER DEFAULT NULL'),
            ('image_note', 'TEXT'),
            ('image_timedelta_posix', 'INTEGER DEFAULT 0'),
            ('image_original_path', 'TEXT'),
            ('image_location_code', 'TEXT'),
            ('party_rowid', 'INTEGER'),
            ('image_metadata_json', 'TEXT'),
        ],
        docstr="""
        First class table used to store image locations and meta-data
        """,
        superkeys=[('image_uuid',)],
        shortname='image',
        extern_tables=['party', 'contributors'],
        dependsmap={
            'contributor_rowid': (
                'contributors',
                ('contributor_rowid',),
                ('contributor_tag',),
            ),
            'party_rowid': ('party', ('party_rowid',), ('party_tag',)),
        },
    )

    db.add_table(
        'imageset_image_relationship',
        [
            ('gsgr_rowid', 'INTEGER PRIMARY KEY'),
            ('image_rowid', 'INTEGER NOT NULL'),
            ('imageset_rowid', 'INTEGER'),
        ],
        docstr="""
        Relationship between imagesets and images (many to many mapping) the
        many-to-many relationship between images and imagesets is encoded here
        imageset_image_relationship stands for imageset-image-pairs.
        """,
        superkeys=[('image_rowid', 'imageset_rowid')],
        relates=('images', 'imagesets'),
        shortname='gsgr',
        dependsmap={
            'image_rowid': ('images', ('image_rowid',), ('image_uuid',)),
            'imageset_rowid': ('imagesets', ('imageset_rowid',), ('imageset_text',)),
        },
    )

    db.add_table(
        'imagesets',
        [
            ('imageset_rowid', 'INTEGER PRIMARY KEY'),
            ('imageset_uuid', 'UUID NOT NULL'),
            ('imageset_text', 'TEXT NOT NULL'),
            ('imageset_occurrence_flag', 'INTEGER DEFAULT 0'),
            ('imageset_note', 'TEXT NOT NULL'),
            ('imageset_start_time_posix', 'INTEGER'),
            ('imageset_end_time_posix', 'INTEGER'),
            ('imageset_gps_lat', 'INTEGER'),
            ('imageset_gps_lon', 'INTEGER'),
            ('imageset_processed_flag', 'INTEGER DEFAULT 0'),
            ('imageset_shipped_flag', 'INTEGER DEFAULT 0'),
            ('imageset_smart_xml_fname', 'TEXT'),
            ('imageset_smart_waypoint_id', 'INTEGER'),
            ('imageset_metadata_json', 'TEXT'),
        ],
        docstr="""
        List of all imagesets. This used to be called the encounter table. It
        represents a group of potentially many individuals seen in a specific
        place at a specific time.
        """,
        superkeys=[('imageset_text',)],
        dependsmap={},
    )

    db.add_table(
        'keys',
        [
            ('lbltype_rowid', 'INTEGER PRIMARY KEY'),
            ('lbltype_text', 'TEXT NOT NULL'),
            ('lbltype_default', 'TEXT NOT NULL'),
        ],
        docstr="""
        List of keys used to define the categories of annotation lables, text
        is for human-readability. The lbltype_default specifies the
        lblannot_value of annotations with a relationship of some
        lbltype_rowid
        """,
        superkeys=[('lbltype_text',)],
    )

    db.add_table(
        'lblannot',
        [
            ('lblannot_rowid', 'INTEGER PRIMARY KEY'),
            ('lblannot_uuid', 'UUID NOT NULL'),
            ('lbltype_rowid', 'INTEGER NOT NULL'),
            ('lblannot_value', 'TEXT NOT NULL'),
            ('lblannot_note', 'TEXT'),
        ],
        docstr="""
        Used to store the labels / attributes of annotations. E.G name,
        species
        """,
        superkeys=[('lbltype_rowid', 'lblannot_value')],
    )

    db.add_table(
        'lblimage',
        [
            ('lblimage_rowid', 'INTEGER PRIMARY KEY'),
            ('lblimage_uuid', 'UUID NOT NULL'),
            ('lbltype_rowid', 'INTEGER NOT NULL'),
            ('lblimage_value', 'TEXT NOT NULL'),
            ('lblimage_note', 'TEXT'),
        ],
        docstr="""
        Used to store the labels (attributes) of images
        """,
        superkeys=[('lbltype_rowid', 'lblimage_value')],
    )

    db.add_table(
        'metadata',
        [
            ('metadata_rowid', 'INTEGER PRIMARY KEY'),
            ('metadata_key', 'TEXT'),
            ('metadata_value', 'TEXT'),
        ],
        docstr="""
        The table that stores permanently all of the metadata about the
        database (tables, etc)
        """,
        superkeys=[('metadata_key',)],
    )

    db.add_table(
        'names',
        [
            ('name_rowid', 'INTEGER PRIMARY KEY'),
            ('name_uuid', 'UUID NOT NULL'),
            ('name_text', 'TEXT NOT NULL'),
            ('name_note', 'TEXT'),
            ('name_temp_flag', 'INTEGER DEFAULT 0'),
            ('name_alias_text', 'TEXT'),
            ('name_sex', 'INTEGER DEFAULT -1'),
            ('name_metadata_json', 'TEXT'),
        ],
        docstr="""
        Stores the individual animal names
        """,
        superkeys=[('name_text',)],
    )

    db.add_table(
        'parts',
        [
            ('part_rowid', 'INTEGER PRIMARY KEY'),
            ('part_uuid', 'UUID NOT NULL'),
            ('annot_rowid', 'INTEGER NOT NULL'),
            ('part_xtl', 'INTEGER NOT NULL'),
            ('part_ytl', 'INTEGER NOT NULL'),
            ('part_width', 'INTEGER NOT NULL'),
            ('part_height', 'INTEGER NOT NULL'),
            ('part_theta', 'REAL DEFAULT 0.0'),
            ('part_num_verts', 'INTEGER NOT NULL'),
            ('part_verts', 'TEXT'),
            ('part_viewpoint', 'TEXT'),
            ('part_detect_confidence', 'REAL DEFAULT -1.0'),
            ('part_toggle_reviewed', 'INTEGER DEFAULT 0'),
            ('part_quality', 'INTEGER'),
            ('part_type', 'TEXT'),
            ('part_note', 'TEXT'),
            ('part_tag_text', 'TEXT'),
            ('part_staged_flag', 'INTEGER DEFAULT 0'),
            ('part_staged_uuid', 'UUID'),
            ('part_staged_user_identity', 'TEXT'),
            ('part_staged_metadata_json', 'TEXT'),
            ('part_metadata_json', 'TEXT'),
            ('part_contour_json', 'TEXT'),
        ],
        docstr="""
        Mainly used to store the geometry of the annotation parts within its
        parent annotation. The one-to-many relationship between annotations
        and parts is encoded here
        """,
        superkeys=[('part_uuid',)],
        shortname='part',
        extern_tables=['annotations'],
        dependsmap={'annot_rowid': ('annotations', ('annot_rowid',), ('annot_uuid',))},
    )

    db.add_table(
        'party',
        [('party_rowid', 'INTEGER PRIMARY KEY'), ('party_tag', 'TEXT NOT NULL')],
        docstr="""
        Serves as a group for contributors
        """,
        superkeys=[('party_tag',)],
    )

    db.add_table(
        'species',
        [
            ('species_rowid', 'INTEGER PRIMARY KEY'),
            ('species_uuid', 'UUID NOT NULL'),
            ('species_text', 'TEXT NOT NULL'),
            ('species_nice', 'TEXT NOT NULL'),
            ('species_code', 'TEXT NOT NULL'),
            ('species_note', 'TEXT'),
            ('species_toggle_enabled', 'INTEGER DEFAULT 1'),
        ],
        docstr="""
        Stores the different animal species
        """,
        superkeys=[('species_text',)],
    )



# -*- coding: utf-8 -*-
"""
Helper module that helps expand parameters for grid search
TODO: move into custom pipe_cfg and annot_cfg modules
"""
import logging
import utool as ut
import sys
import six
import itertools
from wbia.expt import experiment_configs
from wbia.expt import cfghelpers
from wbia.algo import Config
from wbia.init import filter_annots

print, rrr, profile = ut.inject2(__name__)
logger = logging.getLogger('wbia')


def get_varied_pipecfg_lbls(cfgdict_list, pipecfg_list=None):
    if pipecfg_list is None:
        from wbia.algo import Config

        cfg_default_dict = dict(Config.QueryConfig().parse_items())
        cfgx2_lbl = ut.get_varied_cfg_lbls(cfgdict_list, cfg_default_dict)
    else:
        # TODO: group cfgdict by config type and then get varied labels
        cfg_default_dict = None
        cfgx2_lbl = ut.get_varied_cfg_lbls(cfgdict_list, cfg_default_dict)
    return cfgx2_lbl


def get_pipecfg_list(test_cfg_name_list, ibs=None, verbose=None):
    r"""
    Builds a list of varied query configurations. Only custom configs depend on
    an ibs object. The order of the output is not gaurenteed to aggree with
    input order.

    FIXME:
        This breaks if you proot=BC_DTW and ibs is None

    Args:
        test_cfg_name_list (list): list of strs
        ibs (wbia.IBEISController): wbia controller object (optional)

    Returns:
        tuple: (cfg_list, cfgx2_lbl) -
            cfg_list (list): list of config objects
            cfgx2_lbl (list): denotes which parameters are being varied.
                If there is just one config then nothing is varied

    CommandLine:
        python -m wbia get_pipecfg_list:0
        python -m wbia get_pipecfg_list:1 --db humpbacks
        python -m wbia get_pipecfg_list:2

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.expt.experiment_helpers import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='testdb1')
        >>> #test_cfg_name_list = ['best', 'custom', 'custom:sv_on=False']
        >>> #test_cfg_name_list = ['default', 'default:sv_on=False', 'best']
        >>> test_cfg_name_list = ['default', 'default:sv_on=False', 'best']
        >>> # execute function
        >>> (pcfgdict_list, pipecfg_list) = get_pipecfg_list(test_cfg_name_list, ibs)
        >>> # verify results
        >>> assert pipecfg_list[0].sv_cfg.sv_on is True
        >>> assert pipecfg_list[1].sv_cfg.sv_on is False
        >>> pipecfg_lbls = get_varied_pipecfg_lbls(pcfgdict_list)
        >>> result = ('pipecfg_lbls = '+ ut.repr2(pipecfg_lbls))
        >>> print(result)
        pipecfg_lbls = ['default:', 'default:sv_on=False']

    Example1:
        >>> # DISABLE_DOCTEST
        >>> import wbia_flukematch.plugin
        >>> from wbia.expt.experiment_helpers import *  # NOQA
        >>> import wbia
        >>> ibs = wbia.opendb(defaultdb='humpbacks')
        >>> test_cfg_name_list = ['default:pipeline_root=BC_DTW,decision=average,crop_dim_size=[960,500]', 'default:K=[1,4]']
        >>> (pcfgdict_list, pipecfg_list) = get_pipecfg_list(test_cfg_name_list, ibs)
        >>> pipecfg_lbls = get_varied_pipecfg_lbls(pcfgdict_list)
        >>> result = ('pipecfg_lbls = '+ ut.repr2(pipecfg_lbls))
        >>> print(result)
        >>> print_pipe_configs(pcfgdict_list, pipecfg_list)
    """
    if verbose is None:
        verbose = ut.VERBOSE
    if ut.VERBOSE:
        logger.info(
            '[expt_help.get_pipecfg_list] building pipecfg_list using: %s'
            % test_cfg_name_list
        )
    if isinstance(test_cfg_name_list, six.string_types):
        test_cfg_name_list = [test_cfg_name_list]
    _standard_cfg_names = []
    _pcfgdict_list = []

    # HACK: Parse out custom configs first
    for test_cfg_name in test_cfg_name_list:
        if test_cfg_name.startswith('custom:') or test_cfg_name == 'custom':
            logger.info('[expthelpers] Parsing nonstandard custom config')
            assert False, 'custom is no longer supported'
            # if test_cfg_name.startswith('custom:'):
            #    # parse out modifications to custom
            #    cfgstr_list = ':'.join(test_cfg_name.split(':')[1:]).split(',')
            #    augcfgdict = ut.parse_cfgstr_list(cfgstr_list, smartcast=True)
            # else:
            #    augcfgdict = {}
            # # Take the configuration from the wbia object
            # pipe_cfg = ibs.--cfg.query_cfg.deepcopy()
            # # Update with augmented params
            # pipe_cfg.update_query_cfg(**augcfgdict)
            # # Parse out a standard cfgdict
            # cfgdict = dict(pipe_cfg.parse_items())
            # cfgdict['_cfgname'] = 'custom'
            # cfgdict['_cfgstr'] = test_cfg_name
            # _pcfgdict_list.append(cfgdict)
        else:
            _standard_cfg_names.append(test_cfg_name)
    # Handle stanndard configs next
    if len(_standard_cfg_names) > 0:
        # Get parsing information
        # cfg_default_dict = dict(Config.QueryConfig().parse_items())
        # valid_keys = list(cfg_default_dict.keys())
        cfgstr_list = _standard_cfg_names
        named_defaults_dict = ut.dict_subset(
            experiment_configs.__dict__, experiment_configs.TEST_NAMES
        )
        alias_keys = experiment_configs.ALIAS_KEYS
        # Parse standard pipeline cfgstrings
        metadata = {'ibs': ibs}
        dict_comb_list = cfghelpers.parse_cfgstr_list2(
            cfgstr_list,
            named_defaults_dict,
            cfgtype=None,
            alias_keys=alias_keys,
            # Hack out valid keys for humpbacks
            # valid_keys=valid_keys,
            strict=False,
            metadata=metadata,
        )
        # Get varied params (there may be duplicates)
        _pcfgdict_list.extend(ut.flatten(dict_comb_list))

    # Expand cfgdicts into PipelineConfig config objects
    # TODO: respsect different algorithm parameters like flukes
    if ibs is None:
        configclass_list = [Config.QueryConfig] * len(_pcfgdict_list)
    else:
        root_to_config = ibs.depc_annot.configclass_dict.copy()
        from wbia.algo.smk import smk_pipeline

        root_to_config['smk'] = smk_pipeline.SMKRequestConfig
        configclass_list = [
            root_to_config.get(
                _cfgdict.get('pipeline_root', _cfgdict.get('proot', 'vsmany')),
                Config.QueryConfig,
            )
            for _cfgdict in _pcfgdict_list
        ]
    _pipecfg_list = [
        cls(**_cfgdict) for cls, _cfgdict in zip(configclass_list, _pcfgdict_list)
    ]

    # Enforce rule that removes duplicate configs
    # by using feasiblity from wbia.algo.Config
    # TODO: Move this unique finding code to its own function
    # and then move it up one function level so even the custom
    # configs can be uniquified
    _flag_list = ut.flag_unique_items(_pipecfg_list)
    cfgdict_list = ut.compress(_pcfgdict_list, _flag_list)
    pipecfg_list = ut.compress(_pipecfg_list, _flag_list)
    if verbose:
        # for cfg in _pipecfg_list:
        #    logger.info(cfg.get_cfgstr())
        #    logger.info(cfg)
        logger.info(
            '[harn.help] return %d / %d unique pipeline configs from: %r'
            % (len(cfgdict_list), len(_pcfgdict_list), test_cfg_name_list)
        )

    if ut.get_argflag(('--pcfginfo', '--pinfo', '--pipecfginfo')):
        ut.colorprint('Requested PcfgInfo for tests... ', 'red')
        print_pipe_configs(cfgdict_list, pipecfg_list)
        ut.colorprint('Finished Reporting PcfgInfo. Exiting', 'red')
        sys.exit(0)
    return (cfgdict_list, pipecfg_list)


def print_pipe_configs(cfgdict_list, pipecfg_list):
    pipecfg_lbls = get_varied_pipecfg_lbls(cfgdict_list, pipecfg_list)
    # pipecfg_lbls = pipecfg_list
    # assert len(pipecfg_lbls) == len(pipecfg_lbls), 'unequal lens'
    for pcfgx, (pipecfg, lbl) in enumerate(zip(pipecfg_list, pipecfg_lbls)):
        logger.info('+--- %d / %d ===' % (pcfgx, (len(pipecfg_list))))
        ut.colorprint(lbl, 'white')
        logger.info(pipecfg.get_cfgstr())
        logger.info('L___')


def testdata_acfg_names(default_acfg_name_list=['default']):
    flags = ('--aidcfg', '--acfg', '-a')
    acfg_name_list = ut.get_argval(flags, type_=list, default=default_acfg_name_list)
    return acfg_name_list


def parse_acfg_combo_list(acfg_name_list):
    r"""
    Parses the name list into a list of config dicts

    Args:
        acfg_name_list (list): a list of annotation config strings

    Returns:
        list: acfg_combo_list

    CommandLine:
        python -m wbia parse_acfg_combo_list:0
        python -m wbia parse_acfg_combo_list:1
        python -m wbia parse_acfg_combo_list:2

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.expt.experiment_helpers import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import annotation_configs
        >>> acfg_name_list = testdata_acfg_names(['default', 'uncontrolled'])
        >>> acfg_combo_list = parse_acfg_combo_list(acfg_name_list)
        >>> acfg_list = ut.flatten(acfg_combo_list)
        >>> printkw = dict()
        >>> annotation_configs.print_acfg_list(acfg_list, **printkw)
        >>> result = ut.repr2(sorted(acfg_list[0].keys()))
        >>> print(result)
        ['dcfg', 'qcfg']

    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.expt.experiment_helpers import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import annotation_configs
        >>> # double colon :: means expand consistently and force const size
        >>> acfg_name_list = testdata_acfg_names(['unctrl', 'ctrl::unctrl'])
        >>> acfg_name_list = testdata_acfg_names(['unctrl', 'varysize', 'ctrl::unctrl'])
        >>> acfg_name_list = testdata_acfg_names(['unctrl', 'varysize', 'ctrl::varysize', 'ctrl::unctrl'])
        >>> acfg_combo_list = parse_acfg_combo_list(acfg_name_list)
        >>> acfg_list = ut.flatten(acfg_combo_list)
        >>> printkw = dict()
        >>> annotation_configs.print_acfg_list(acfg_list, **printkw)
    """
    from wbia.expt import annotation_configs

    named_defaults_dict = ut.dict_take(
        annotation_configs.__dict__, annotation_configs.TEST_NAMES
    )
    named_qcfg_defaults = dict(
        zip(
            annotation_configs.TEST_NAMES,
            ut.get_list_column(named_defaults_dict, 'qcfg'),
        )
    )
    named_dcfg_defaults = dict(
        zip(
            annotation_configs.TEST_NAMES,
            ut.get_list_column(named_defaults_dict, 'dcfg'),
        )
    )
    alias_keys = annotation_configs.ALIAS_KEYS
    # need to have the cfgstr_lists be the same for query and database so they
    # can be combined properly for now

    # Apply this flag to any case joined with ::
    special_join_dict = {'force_const_size': True}

    # Parse Query Annot Config
    nested_qcfg_combo_list = cfghelpers.parse_cfgstr_list2(
        cfgstr_list=acfg_name_list,
        named_defaults_dict=named_qcfg_defaults,
        cfgtype='qcfg',
        alias_keys=alias_keys,
        expand_nested=False,
        special_join_dict=special_join_dict,
        is_nestedcfgtype=True,
    )

    # Parse Data Annot Config
    nested_dcfg_combo_list = cfghelpers.parse_cfgstr_list2(
        cfgstr_list=acfg_name_list,
        named_defaults_dict=named_dcfg_defaults,
        cfgtype='dcfg',
        alias_keys=alias_keys,
        expand_nested=False,
        special_join_dict=special_join_dict,
        is_nestedcfgtype=True,
    )

    acfg_combo_list = []
    for nested_qcfg_combo, nested_dcfg_combo in zip(
        nested_qcfg_combo_list, nested_dcfg_combo_list
    ):
        acfg_combo = []
        # Only the inner nested combos are combinatorial
        for qcfg_combo, dcfg_combo in zip(nested_qcfg_combo, nested_dcfg_combo):
            _combo = [
                dict([('qcfg', qcfg), ('dcfg', dcfg)])
                for qcfg, dcfg in list(itertools.product(qcfg_combo, dcfg_combo))
            ]
            acfg_combo.extend(_combo)
        acfg_combo_list.append(acfg_combo)
    return acfg_combo_list


def filter_duplicate_acfgs(expanded_aids_list, acfg_list, acfg_name_list, verbose=None):
    """
    Removes configs with the same expanded aids list

    CommandLine:
        # The following will trigger this function:
        wbia -m wbia get_annotcfg_list:0 -a timectrl timectrl:view=left --db PZ_MTEST

    """
    from wbia.expt import annotation_configs

    if verbose is None:
        verbose = ut.VERBOSE
    acfg_list_ = []
    expanded_aids_list_ = []
    seen_ = ut.ddict(list)
    for acfg, (qaids, daids) in zip(acfg_list, expanded_aids_list):
        key = (ut.hashstr_arr27(qaids, 'qaids'), ut.hashstr_arr27(daids, 'daids'))
        if key in seen_:
            seen_[key].append(acfg)
            continue
        else:
            seen_[key].append(acfg)
            expanded_aids_list_.append((qaids, daids))
            acfg_list_.append(acfg)
    if verbose:
        duplicate_configs = dict(
            [(key_, val_) for key_, val_ in seen_.items() if len(val_) > 1]
        )
        if len(duplicate_configs) > 0:
            logger.info('The following configs produced duplicate annnotation configs')
            for key, val in duplicate_configs.items():
                # Print the difference between the duplicate configs
                _tup = annotation_configs.compress_acfg_list_for_printing(val)
                nonvaried_compressed_dict, varied_compressed_dict_list = _tup
                logger.info('+--')
                logger.info('key = %r' % (key,))
                logger.info(
                    'duplicate_varied_cfgs = %s'
                    % (ut.repr2(varied_compressed_dict_list),)
                )
                logger.info(
                    'duplicate_nonvaried_cfgs = %s'
                    % (ut.repr2(nonvaried_compressed_dict),)
                )
                logger.info('L__')

        if verbose >= 1:
            logger.info(
                '[harn.help] parsed %d / %d unique annot configs'
                % (len(acfg_list_), len(acfg_list))
            )
        if verbose > 2:
            logger.info('[harn.help] parsed from: %r' % (acfg_name_list,))
    return expanded_aids_list_, acfg_list_


def get_annotcfg_list(
    ibs,
    acfg_name_list,
    filter_dups=True,
    qaid_override=None,
    daid_override=None,
    initial_aids=None,
    use_cache=None,
    verbose=None,
):
    r"""
    For now can only specify one acfg name list

    TODO: move to filter_annots

    Args:
        annot_cfg_name_list (list):

    CommandLine:
        python -m wbia get_annotcfg_list:0
        python -m wbia get_annotcfg_list:1
        python -m wbia get_annotcfg_list:2

        wbia get_annotcfg_list:0 --ainfo
        wbia get_annotcfg_list:0 --db NNP_Master3 -a viewpoint_compare --nocache-aid --verbtd
        wbia get_annotcfg_list:0 --db PZ_ViewPoints -a viewpoint_compare --nocache-aid --verbtd
        wbia get_annotcfg_list:0 --db PZ_MTEST -a unctrl ctrl::unctrl --ainfo --nocache-aid
        wbia get_annotcfg_list:0 --db testdb1 -a : --ainfo --nocache-aid
        wbia get_annotcfg_list:0 --db Oxford -a :qhas_any=query --ainfo --nocache-aid
        wbia get_annotcfg_list:0 --db Oxford -a :qhas_any=query,dhas_any=distractor --ainfo --nocache-aid

    Example0:
        >>> # DISABLE_DOCTEST
        >>> from wbia.expt.experiment_helpers import *  # NOQA
        >>> import wbia
        >>> from wbia.expt import annotation_configs
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> filter_dups = not ut.get_argflag('--nofilter-dups')
        >>> acfg_name_list = testdata_acfg_names()
        >>> _tup = get_annotcfg_list(ibs, acfg_name_list, filter_dups)
        >>> acfg_list, expanded_aids_list = _tup
        >>> print('\n PRINTING TEST RESULTS')
        >>> result = ut.repr2(acfg_list, nl=3)
        >>> print('\n')
        >>> #statskw = ut.parse_func_kwarg_keys(ibs.get_annot_stats_dict, with_vals=False)
        >>> printkw = dict(combined=True, per_name_vpedge=None,
        >>>                per_qual=False, per_vp=False, case_tag_hist=False)
        >>> annotation_configs.print_acfg_list(
        >>>     acfg_list, expanded_aids_list, ibs, **printkw)


    Example:
        >>> # ENABLE_DOCTEST
        >>> from wbia.expt.experiment_helpers import *  # NOQA
        >>> import wbia
        >>> from wbia.init import main_helpers
        >>> from wbia.expt import annotation_configs
        >>> ibs = wbia.opendb(defaultdb='PZ_MTEST')
        >>> aids = ibs.get_valid_aids()
        >>> main_helpers.monkeypatch_encounters(ibs, aids, days=50)
        >>> a = ['default:crossval_enc=True,require_timestamp=True']
        >>> acfg_name_list = testdata_acfg_names(a)
        >>> acfg_list, expanded_aids_list = get_annotcfg_list(ibs, acfg_name_list)
        >>> annotation_configs.print_acfg_list(acfg_list, expanded_aids_list)
        >>> # Restore state
        >>> main_helpers.unmonkeypatch_encounters(ibs)
    """
    if ut.VERBOSE:
        logger.info('[harn.help] building acfg_list using %r' % (acfg_name_list,))
    from wbia.expt import annotation_configs

    acfg_combo_list = parse_acfg_combo_list(acfg_name_list)

    # acfg_slice = ut.get_argval('--acfg_slice', type_=slice, default=None)
    # HACK: Sliceing happens before expansion (dependenceis get)
    combo_slice = ut.get_argval(
        '--combo_slice', type_='fuzzy_subset', default=slice(None)
    )
    acfg_combo_list = [
        ut.take(acfg_combo_, combo_slice) for acfg_combo_ in acfg_combo_list
    ]

    if ut.get_argflag('--consistent'):
        # Expand everything as one consistent annot list
        acfg_combo_list = [ut.flatten(acfg_combo_list)]

    # + --- Do Parsing ---
    expanded_aids_combo_list = [
        filter_annots.expand_acfgs_consistently(
            ibs,
            acfg_combo_,
            initial_aids=initial_aids,
            use_cache=use_cache,
            verbose=verbose,
            base=base,
        )
        for base, acfg_combo_ in enumerate(acfg_combo_list)
    ]
    expanded_aids_combo_flag_list = ut.flatten(expanded_aids_combo_list)
    acfg_list = ut.get_list_column(expanded_aids_combo_flag_list, 0)
    expanded_aids_list = ut.get_list_column(expanded_aids_combo_flag_list, 1)
    # L___

    # Slicing happens after expansion (but the labels get screwed up)
    acfg_slice = ut.get_argval('--acfg_slice', type_='fuzzy_subset', default=None)
    if acfg_slice is not None:
        acfg_list = ut.take(acfg_list, acfg_slice)
        expanded_aids_list = ut.take(expanded_aids_list, acfg_slice)

    # + --- Hack: Override qaids ---
    _qaids = ut.get_argval(
        ('--qaid', '--qaid-override'), type_=list, default=qaid_override
    )
    if _qaids is not None:
        expanded_aids_list = [(_qaids, daids) for qaids, daids in expanded_aids_list]
    # more hack for daids
    _daids = ut.get_argval(
        ('--daids-override', '--daid-override'), type_=list, default=daid_override
    )
    if _daids is not None:
        expanded_aids_list = [(qaids, _daids) for qaids, daids in expanded_aids_list]
    # L___

    if filter_dups:
        expanded_aids_list, acfg_list = filter_duplicate_acfgs(
            expanded_aids_list, acfg_list, acfg_name_list
        )

    if ut.get_argflag(
        ('--acfginfo', '--ainfo', '--aidcfginfo', '--print-acfg', '--printacfg')
    ):
        ut.colorprint('[experiment_helpers] Requested AcfgInfo ... ', 'red')
        logger.info('combo_slice = %r' % (combo_slice,))
        logger.info('acfg_slice = %r' % (acfg_slice,))
        annotation_configs.print_acfg_list(acfg_list, expanded_aids_list, ibs)
        ut.colorprint('[experiment_helpers] exiting due to AcfgInfo info request', 'red')
        sys.exit(0)

    return acfg_list, expanded_aids_list



# -*- coding: utf-8 -*-
# Autogenerated on 12:39:11 2016/10/13
# flake8: noqa
import logging
from wbia.algo.hots import _pipeline_helpers
from wbia.algo.hots import chip_match
from wbia.algo.hots import exceptions
from wbia.algo.hots import hstypes
from wbia.algo.hots import match_chips4
from wbia.algo.hots import name_scoring
from wbia.algo.hots import neighbor_index
from wbia.algo.hots import neighbor_index_cache
from wbia.algo.hots import nn_weights
from wbia.algo.hots import old_chip_match
from wbia.algo.hots import pipeline
from wbia.algo.hots import query_params
from wbia.algo.hots import query_request
from wbia.algo.hots import scoring
import utool

print, rrr, profile = utool.inject2(__name__, '[wbia.algo.hots]')
logger = logging.getLogger('wbia')


def reassign_submodule_attributes(verbose=True):
    """
    why reloading all the modules doesnt do this I don't know
    """
    import sys

    if verbose and '--quiet' not in sys.argv:
        print('dev reimport')
    # Self import
    import wbia.algo.hots

    # Implicit reassignment.
    seen_ = set([])
    for tup in IMPORT_TUPLES:
        if len(tup) > 2 and tup[2]:
            continue  # dont import package names
        submodname, fromimports = tup[0:2]
        submod = getattr(wbia.algo.hots, submodname)
        for attr in dir(submod):
            if attr.startswith('_'):
                continue
            if attr in seen_:
                # This just holds off bad behavior
                # but it does mimic normal util_import behavior
                # which is good
                continue
            seen_.add(attr)
            setattr(wbia.algo.hots, attr, getattr(submod, attr))


def reload_subs(verbose=True):
    """ Reloads wbia.algo.hots and submodules """
    if verbose:
        print('Reloading submodules')
    rrr(verbose=verbose)

    def wrap_fbrrr(mod):
        def fbrrr(*args, **kwargs):
            """ fallback reload """
            if verbose:
                print('No fallback relaod for mod=%r' % (mod,))
            # Breaks ut.Pref (which should be depricated anyway)
            # import imp
            # imp.reload(mod)

        return fbrrr

    def get_rrr(mod):
        if hasattr(mod, 'rrr'):
            return mod.rrr
        else:
            return wrap_fbrrr(mod)

    def get_reload_subs(mod):
        return getattr(mod, 'reload_subs', wrap_fbrrr(mod))

    get_rrr(_pipeline_helpers)(verbose=verbose)
    get_rrr(chip_match)(verbose=verbose)
    get_rrr(exceptions)(verbose=verbose)
    get_rrr(hstypes)(verbose=verbose)
    get_rrr(match_chips4)(verbose=verbose)
    get_rrr(name_scoring)(verbose=verbose)
    get_rrr(neighbor_index)(verbose=verbose)
    get_rrr(neighbor_index_cache)(verbose=verbose)
    get_rrr(nn_weights)(verbose=verbose)
    get_rrr(old_chip_match)(verbose=verbose)
    get_rrr(pipeline)(verbose=verbose)
    get_rrr(query_params)(verbose=verbose)
    get_rrr(query_request)(verbose=verbose)
    get_rrr(scoring)(verbose=verbose)
    rrr(verbose=verbose)
    try:
        # hackish way of propogating up the new reloaded submodule attributes
        reassign_submodule_attributes(verbose=verbose)
    except Exception as ex:
        print(ex)


rrrr = reload_subs

IMPORT_TUPLES = [
    ('_pipeline_helpers', None),
    ('chip_match', None),
    ('exceptions', None),
    ('hstypes', None),
    ('match_chips4', None),
    ('name_scoring', None),
    ('neighbor_index', None),
    ('neighbor_index_cache', None),
    ('nn_weights', None),
    ('old_chip_match', None),
    ('pipeline', None),
    ('query_params', None),
    ('query_request', None),
    ('scoring', None),
]
"""
Regen Command:
    cd /home/joncrall/code/wbia/wbia/algo/hots
    makeinit.py --modname=wbia.algo.hots
"""
